"""
implementation of log sv model calibration to time series
need package option_chain_analytics with BTC and ETH data
"""
import pandas as pd
import numpy as np
import matplotlib.pyplot as plt
import seaborn as sns
import qis as qis
from typing import Dict, Tuple, Optional
from scipy.optimize import minimize_scalar
from enum import Enum

# analytics
from stochvolmodels import (OptionChain, GmmParams, GmmPricer,
                            generate_vol_chain_np, sample_option_chain_at_times)


# chain data
from option_chain_analytics import OptionsDataDFs, create_chain_from_from_options_dfs
from option_chain_analytics.ts_loaders import ts_data_loader_wrapper

from stochvolmodels.pricers.gmm_pricer import plot_gmm_pdfs

FIGSIZE = (16, 4.5)


def plot_calibrated_gmm_model_per_slice(option_chain: OptionChain,
                                        n_mixtures: int = 4
                                        ) -> Dict[str, GmmParams]:
    """
    plot implied vols vs market from gmm model
    """
    gmm_pricer = GmmPricer()
    calibrated_params = {}
    with sns.axes_style("darkgrid"):
        fig, axs = plt.subplots(1, len(option_chain.ids), figsize=(18, 10), tight_layout=True)
        if len(option_chain.ids) == 1:
            axs = [axs]
    for idx, ids_ in enumerate(option_chain.ids):
        option_chain0 = OptionChain.get_slices_as_chain(option_chain, ids=[ids_])
        calibrated_params_t = gmm_pricer.calibrate_model_params_to_chain_slice(option_chain=option_chain0,
                                                                               n_mixtures=n_mixtures)
        calibrated_params[ids_] = calibrated_params_t
        gmm_pricer.plot_model_ivols_vs_bid_ask(option_chain=option_chain0, params=calibrated_params_t,
                                               axs=[axs[idx]])
    return calibrated_params


def plot_riskpremia_pdfs(params: GmmParams,
                         kappa: float = 3.0,
                         nstdev: float = 6.0,
                         title: str = None,
                         ax: plt.Subplot = None
                         ) -> None:
    stdev = nstdev*params.get_get_avg_vol()*np.sqrt(params.ttm)
    x = np.linspace(-stdev, stdev, 2000)
    risk_neutral_pdf = params.compute_pdf(x=x)
    risk_neutral_pdf = risk_neutral_pdf / np.sum(risk_neutral_pdf)
    kernel = np.exp(kappa*x)
    statistical_pdf = kernel*risk_neutral_pdf
    statistical_pdf = statistical_pdf / np.sum(statistical_pdf)
    print(f"forward_q={np.sum(risk_neutral_pdf*np.exp(x))}, forward_p={np.sum(statistical_pdf*np.exp(x))}")

    risk_neutral_pdf = pd.Series(risk_neutral_pdf, index=x, name='risk neutral')
    statistical_pdf = pd.Series(statistical_pdf, index=x, name='statistical')
    df = pd.concat([risk_neutral_pdf, statistical_pdf], axis=1)
    kwargs = dict(fontsize=14, framealpha=0.80)
    qis.plot_line(df=df,
                  yvar_format=None,
                  y_limits=(0.0, None),
                  title=title,
                  xlabel='log-return',
                  ax=ax,
                  **kwargs)
    ax.axes.get_yaxis().set_visible(False)


def compute_risk_premia(params: GmmParams, kappa: float = 3.0) -> float:
    alpha_i = params.gmm_mus * params.ttm
    v_i = np.square(params.gmm_vols) * params.ttm
    gamma_i = kappa*alpha_i+0.5*np.square(kappa)*v_i
    weights_i = params.gmm_weights*np.exp(gamma_i)
    weights_i = weights_i / np.sum(weights_i)
    rp = (np.sum(weights_i*np.exp(alpha_i+(kappa+0.5)*v_i)) - 1.0) / params.ttm
    return rp


def fit_kappa(returns: pd.Series, span: int = None) -> float:
    """
    fit kappa from returns
    """
    x = returns.to_numpy()
    if span is not None:
        weights = qis.compute_expanding_power(n=len(x), power_lambda=1.0-2.0/(span+1.0), reverse_columns=True)
    else:
        weights = None

    def f(kappa: float):  # sqyute of f(kappa) = 0
        if weights is not None:
            res = np.sum(weights * np.exp(-kappa * x) * (np.exp(x) - 1.0))
        else:
            res = np.sum(np.exp(-kappa*x)*(np.exp(x)-1.0))
        return np.square(res)

    options = {'disp': False, 'maxiter': 300}
    res = minimize_scalar(f, bounds=(-10.0, 10.0), options=options, tol=1e-12)
    return res.x


def fit_rolling_kappa(prices: pd.Series,
                      span: int = None,
                      reb_freq: str = 'M-FRI',
                      hour_offset: int = 8
                      ):
    """
    fit rolling kappa with schedule at reb_freq
    """
    returns = qis.to_returns(prices=prices, is_log_returns=True, drop_first=True)
    rebalancing_schedule = qis.generate_dates_schedule(time_period=qis.get_time_period(returns, tz='UTC'),
                                                       freq=reb_freq,
                                                       hour_offset=hour_offset)
    kappas = {}
    for date in rebalancing_schedule:
        kappas[date] = fit_kappa(returns=returns[:date], span=span)
    kappas = pd.Series(kappas)
    return kappas


def calibrate_time_series_of_risk_premia(options_data_dfs: OptionsDataDFs,
                                         time_period: qis.TimePeriod,
                                         freq: str = 'W-FRI',
                                         hour_offset: int = 8,
                                         span: int = 12 * 30 * 24,
                                         days_map: Dict[str, int] = {'1w': 7, '1m': 21, '1q': 60},
                                         delta_bounds: Tuple[Optional[float], Optional[float]] = (-0.05, 0.05),
                                         n_mixtures: int = 4
                                         ) -> pd.DataFrame:
    """
    generate chains at given freqs and report calibration stats
    """
    option_chains = sample_option_chain_at_times(options_data_dfs=options_data_dfs,
                                                 time_period=time_period,
                                                 freq=freq,
                                                 hour_offset=hour_offset,
                                                 days_map=days_map,
                                                 delta_bounds=delta_bounds)
    chain_dates = pd.DatetimeIndex(option_chains.keys())
    kappas = fit_rolling_kappa(prices=options_data_dfs.get_spot_data()['close'],
                               span=span,
                               reb_freq=freq,
                               hour_offset=hour_offset)
    kappas_at_chain_dates = kappas.reindex(index=chain_dates).ffill()

    risk_premias = {}
    gmm_pricer = GmmPricer()
    for date, option_chain in option_chains.items():
        calibrated_params = gmm_pricer.calibrate_model_params_to_chain(option_chain=option_chain, n_mixtures=n_mixtures)
        print(calibrated_params)
        kappa = kappas_at_chain_dates[date]
        risk_premia_by_term = {'kappa': kappa}
        for key, params in calibrated_params.items():
            # key is "term: expiry"
            risk_premia = compute_risk_premia(params=params, kappa=kappa)
            risk_premia_by_term[key.split(':')[0]] = risk_premia
        risk_premias[date] = risk_premia_by_term
    risk_premias = pd.DataFrame.from_dict(risk_premias, orient='index')
    return risk_premias


def plot_rolling_kappa(prices: pd.Series,
                       span: int = 12*30*24,
                       reb_freq: str = 'W-FRI'
                       ) -> plt.Figure:
    kappas = fit_rolling_kappa(prices=prices,
                               span=span,
                               reb_freq=reb_freq)
    kappas = kappas.iloc[26:, ].rename('lambda')

    returns = prices.pct_change()
    rolling_returns = returns.rolling(30*24).sum().rename('returns')
    # rolling_returns = prices.reindex(index=kappas.index).pct_change().rename('returns')
    df = pd.concat([rolling_returns, kappas], axis=1).dropna()

    with sns.axes_style("darkgrid"):
        kwargs = dict(fontsize=14)
        fig, axs = plt.subplots(2, 1, figsize=(16, 17), tight_layout=True)
        qis.plot_time_series(df=kappas,
                             title='Lamda time series',
                             date_format='%b-%y',
                             ax=axs[0],
                             **kwargs)

        # fig2, ax = plt.subplots(1, 1, figsize=(16, 7))
        qis.df_boxplot_by_classification_var(df=df, x='returns', y='lambda', num_buckets=4,
                                             x_hue_name='rolling monthly return',
                                             title='Quantiles of lambda conditional on quantiles of rolling monthly returns',
                                             xvar_format='{:.0%}',
                                             ax=axs[1],
                                             **kwargs)
    qis.plot_classification_scatter(df=df, x='returns', y='lambda', num_buckets=4)

    return fig


def strategy_analysis(options_risk_premia: pd.Series,
                      strategy_nav: pd.Series,
                      rp_bound: float = 10.0
                      ) -> None:
    # get strat returns aligned with riskpremia
    nav = strategy_nav.reindex(index=options_risk_premia.index)
    returns = qis.to_returns(nav)

    options_risk_premia = pd.Series(np.clip(options_risk_premia, a_min=None, a_max=rp_bound),
                                    index=options_risk_premia.index, name=options_risk_premia.name)
    # shift risk-premia as predictor
    df = pd.concat([options_risk_premia.shift(1), returns], axis=1).dropna()

    # filter out rp in excess of 10 for scatters
    df1 = df.loc[np.less(np.abs(df[options_risk_premia.name]), rp_bound), :]

    # scatter first
    qis.plot_scatter(df=df1)

    x = str(options_risk_premia.name)
    y = str(strategy_nav.name)

    kwargs = dict(fit_intercept=False)
    qis.plot_classification_scatter(df=df1, x=x, y=y, num_buckets=3, **kwargs)
    qis.plot_classification_scatter(df=df1, x=x, y=y, num_buckets=4, **kwargs)
    qis.plot_classification_scatter(df=df1, x=x, y=y, num_buckets=5, **kwargs)
    qis.plot_classification_scatter(df=df1, x=x, y=y, num_buckets=6, **kwargs)

    qis.df_boxplot_by_classification_var(df=df1, x=x, y=y, num_buckets=3)
    qis.df_boxplot_by_classification_var(df=df1, x=x, y=y, num_buckets=4)
    qis.df_boxplot_by_classification_var(df=df1, x=x, y=y, num_buckets=5)
    qis.df_boxplot_by_classification_var(df=df1, x=x, y=y, num_buckets=6)

    # compute strategy full and filtered returns
    returns_full = df.iloc[:, 1].rename('full')

    rp = df.iloc[:, 0]
    qs = 1.0 - np.array([1.0 / 4.0, 1.0 / 5.0, 1.0 / 6.0, 1.0 / 8.0])
    rp_quantiles = np.nanquantile(a=rp.to_numpy(), q=qs)
    rp_quantiles_ = {}
    returns_cut = {}
    for idx, q in enumerate(qs):
        key = f"quantile-{idx+4}"
        running_q = df.iloc[:, 0].rolling(156, min_periods=4).quantile(q)
        rp_quantiles_[key] = running_q
        returns_cut[key] = df.iloc[:, 1].where(df.iloc[:, 0] < running_q.to_numpy(), other=0.0)
    rp_quantiles_ = pd.DataFrame.from_dict(rp_quantiles_, orient='columns')
    print(rp_quantiles_)

    # for idx, qcut in enumerate(rp_quantiles):
    #    key = f"quantile-{idx+4}"
    #    returns_cut[key] = df.iloc[:, 1].where(df.iloc[:, 0] < qcut, other=0.0)

    returns_cut = pd.DataFrame.from_dict(returns_cut, orient='columns')

    all = pd.concat([returns_full, returns_cut], axis=1).fillna(0.0)
    navs = qis.returns_to_nav(returns=all)
    qis.plot_prices(navs, perf_params=qis.PerfParams(freq='W-FRI'))


def final_figures(options_risk_premia: pd.Series,
                  strategy_nav: pd.Series,
                  rp_bound: float = 10.0
                  ) -> Tuple[plt.Figure, plt.Figure]:
    # get strat returns aligned with riskpremia
    nav = strategy_nav.reindex(index=options_risk_premia.index)
    returns = qis.to_returns(nav)

    options_risk_premia = pd.Series(np.clip(options_risk_premia, a_min=None, a_max=rp_bound),
                                    index=options_risk_premia.index, name=options_risk_premia.name)
    # shift risk-premia as predictor
    df = pd.concat([options_risk_premia.shift(1), returns], axis=1).dropna()

    # filter out rp in excess of 10 for scatters
    df1 = df.loc[np.less(np.abs(df[options_risk_premia.name]), rp_bound), :]

    x = str(options_risk_premia.name)
    y = str(strategy_nav.name)

    with sns.axes_style("darkgrid"):
        kwargs = dict(fontsize=14)
        fig1, axs = plt.subplots(2, 1, figsize=(16, 17), tight_layout=True)
        qis.plot_time_series(df=options_risk_premia,
                             title='Time series of risk-premia inferred from weekly options',
                             date_format='%b-%y',
                             ax=axs[0],
                             **kwargs)

        qis.df_boxplot_by_classification_var(df=df, x=x, y=y, num_buckets=4,
                                             x_hue_name='risk-premia on previous week',
                                             title='Quantiles of straddle weekly returns conditional on previous week risk-premia',
                                             xvar_format='{:.0%}',
                                             ax=axs[1],
                                             **kwargs)

        # compute strategy full and filtered returns
        returns_full = df.iloc[:, 1].rename('Unconstrained')

        rp = df.iloc[:, 0]
        qs = 1.0 - np.array([1.0 / 4.0, 1.0 / 5.0, 1.0 / 6.0, 1.0 / 8.0])
        rp_quantiles = np.nanquantile(a=rp.to_numpy(), q=qs)
        rp_quantiles_ = {}
        returns_cut = {}
        for idx, q in enumerate(qs):
            key = f"Filtered quantile-{idx + 4}"
            running_q = df.iloc[:, 0].rolling(156, min_periods=4).quantile(q)
            rp_quantiles_[key] = running_q
            returns_cut[key] = df.iloc[:, 1].where(df.iloc[:, 0] < running_q.to_numpy(), other=0.0)
        rp_quantiles_ = pd.DataFrame.from_dict(rp_quantiles_, orient='columns')

        # for idx, qcut in enumerate(rp_quantiles):
        #    key = f"quantile-{idx+4}"
        #    returns_cut[key] = df.iloc[:, 1].where(df.iloc[:, 0] < qcut, other=0.0)

        returns_cut = pd.DataFrame.from_dict(returns_cut, orient='columns')

        all = pd.concat([returns_full, returns_cut], axis=1).fillna(0.0)
        navs = qis.returns_to_nav(returns=all)

        fig2, ax = plt.subplots(1, 1, figsize=(16, 17), tight_layout=True)
        qis.plot_prices(navs,
                        perf_params=qis.PerfParams(freq='W-FRI'),
                        ax=ax,
                        **kwargs)

    return fig1, fig2


class UnitTests(Enum):
    FIT_GMM_MODEL = 1
    FIGURE1_RISK_PREMIA = 3
    FIGURE2_PLOT_PDFS = 2
    FIT_KAPPA = 4
    FIGURE3_PLOT_ROLLING_KAPPA = 5
    FIT_TIME_SERIES_RISK_PREMIA = 6
    STRATEGY_ANALYSIS = 7
    FINAL_FIGURES = 8


def run_unit_test(unit_test: UnitTests):

    local_path = 'C://Users//artur//OneDrive//My Papers//MyPresentations//Risk-premia. Zurich. Feb 2024//Figures//'

    pd.set_option('display.max_columns', 1000)

    ticker = 'BTC'  # BTC, ETH
    value_time = pd.Timestamp('2019-06-28 10:00:00+00:00')
    value_time = pd.Timestamp('2022-12-30 10:00:00+00:00')
    value_time = pd.Timestamp('2023-06-30 10:00:00+00:00')
    value_time = pd.Timestamp('2024-01-03 10:00:00+00:00')
    value_time = pd.Timestamp('2024-01-04 10:00:00+00:00')
    value_time = pd.Timestamp('2021-01-08 10:00:00+00:00')
    value_time = pd.Timestamp('2024-01-05 10:00:00+00:00')
    # value_time = pd.Timestamp('2024-01-11 10:00:00+00:00')

    # chain data here
    options_data_dfs = OptionsDataDFs(**ts_data_loader_wrapper(ticker=ticker))
    chain = create_chain_from_from_options_dfs(options_data_dfs=options_data_dfs, value_time=value_time)

    option_chain = generate_vol_chain_np(chain=chain,
                                         value_time=value_time,
                                         # days_map={'1w': 7, '2w': 14, '1m': 21},
                                         days_map={'1w': 7, '1m': 21, '1q': 60},
                                         delta_bounds=(-0.05, 0.05),
                                         #delta_bounds=(-0.1, 0.1),
                                         is_filtered=True)

    # option_chain.print()
    # print(option_chain0)

    if unit_test == UnitTests.FIT_GMM_MODEL:
        gmm_pricer = GmmPricer()
        for ids in option_chain.ids:
            option_chain0 = OptionChain.get_slices_as_chain(option_chain, ids=ids)
            params = gmm_pricer.calibrate_model_params_to_chain_slice(option_chain=option_chain0,
                                                                                 n_mixtures=4)
            print(f"{ids}: {params}")
            plot_gmm_pdfs(params=params, option_chain0=option_chain0)

    elif unit_test == UnitTests.FIGURE1_RISK_PREMIA:
        option_chain0 = OptionChain.get_slices_as_chain(option_chain, ids=[option_chain.ids[0]])
        calibrated_params = plot_calibrated_gmm_model_per_slice(option_chain=option_chain0,
                                                                n_mixtures=4)
        print(calibrated_params)
        params = calibrated_params[list(calibrated_params.keys())[0]]

        with sns.axes_style("darkgrid"):
            fig, axs = plt.subplots(1, 3, figsize=FIGSIZE)
            plot_riskpremia_pdfs(params=params, kappa=3.0, title='lambda=3.0', ax=axs[0])
            plot_riskpremia_pdfs(params=params, kappa=0.0, title='lambda=0.0', ax=axs[1])
            plot_riskpremia_pdfs(params=params, kappa=-3.0, title='lambda=-3.0', ax=axs[2])
            qis.set_suptitle(fig, title='Risk-neutral vs statistical PDFs under risk-preferences', fontsize=16)

        qis.save_fig(fig=fig, file_name='figure1_rp_pdf', local_path=local_path)
        # options_rp = compute_risk_premia(params=params, kappa=3.0)
        # print(f"options_rp={options_rp}")

    elif unit_test == UnitTests.FIGURE2_PLOT_PDFS:
        gmm_pricer = GmmPricer()
        # 1m: 23Feb2024
        option_chain0 = OptionChain.get_slices_as_chain(option_chain, ids=[option_chain.ids[1]])

        params3 = gmm_pricer.calibrate_model_params_to_chain_slice(option_chain=option_chain0, n_mixtures=3)
        print(params3)
        params4 = gmm_pricer.calibrate_model_params_to_chain_slice(option_chain=option_chain0, n_mixtures=4)
        print(params4)

        with sns.axes_style("darkgrid"):
            fig, axs = plt.subplots(2, 2, figsize=(16, 9))

        plot_gmm_pdfs(params=params3, option_chain0=option_chain0, axs=axs[0, :])
        plot_gmm_pdfs(params=params4, option_chain0=option_chain0, axs=axs[1, :])

        qis.set_suptitle(fig=fig, title='3 state (top) and 4 state (bottom) GMM', fontsize=16)
        qis.save_fig(fig=fig, file_name='figure2_state_pdfs', local_path=local_path)

    elif unit_test == UnitTests.FIT_KAPPA:
        prices = options_data_dfs.get_spot_data()['close']
        print(prices)
        returns = qis.to_returns(prices=prices, is_log_returns=True, drop_first=True)
        kappa = fit_kappa(returns=returns, span=30*24)
        print(kappa)

    elif unit_test == UnitTests.FIGURE3_PLOT_ROLLING_KAPPA:

        fig = plot_rolling_kappa(prices=options_data_dfs.get_spot_data()['close'],
                                 span=12*30*24,
                                 reb_freq='W-FRI')
        qis.set_suptitle(fig=fig, title='Risk-preference lambda', fontsize=16)
        qis.save_fig(fig=fig, file_name='figure3_lambda_ts', local_path=local_path)

    elif unit_test == UnitTests.FIT_TIME_SERIES_RISK_PREMIA:
        time_period = qis.TimePeriod('28Feb2020', '02Mar2024', tz='UTC')

        risk_premias = calibrate_time_series_of_risk_premia(options_data_dfs=options_data_dfs,
                                                            time_period=time_period,
                                                            freq='W-FRI',
                                                            hour_offset=10,
                                                            days_map={'1w': 7, '1m': 21, '1q': 60},
                                                            span=12 * 30 * 24,
                                                            delta_bounds=(-0.1, 0.1),
                                                            n_mixtures=4)
        qis.plot_time_series(df=risk_premias)

        qis.save_df_to_csv(df=risk_premias, file_name='btc_riskpremia_12m_4',
                           local_path="C://Users//artur//OneDrive//analytics//resources//")

    elif unit_test == UnitTests.STRATEGY_ANALYSIS:
        options_risk_premia = qis.load_df_from_csv(file_name='btc_riskpremia_12m_3',
                                                   local_path="C://Users//artur//OneDrive//analytics//resources//")
        print(options_risk_premia)
        navs = qis.load_df_from_csv(file_name='btc_straddle',
                                           local_path="C://Users//artur//OneDrive//analytics//resources//")
        print(navs)
        options_risk_premia = options_risk_premia['1w'].rename('risk-premia')
        # options_risk_premia = np.subtract(options_risk_premia['1w'], options_risk_premia['kappa']).rename('risk-premia')
        # options_risk_premia = options_risk_premia['kappa'].rename('risk-premia')
        # strategy_nav = navs['BTC'].rename('BTC')
        strategy_nav = navs['portfolio_nav'].rename('Short Straddle')

        strategy_analysis(options_risk_premia=options_risk_premia,
                          strategy_nav=strategy_nav,
                          rp_bound=5.0)

    elif unit_test == UnitTests.FINAL_FIGURES:
        options_risk_premia = qis.load_df_from_csv(file_name='btc_riskpremia_12m_3',
                                                   local_path="C://Users//artur//OneDrive//analytics//resources//")
        print(options_risk_premia)
        navs = qis.load_df_from_csv(file_name='btc_straddle',
                                           local_path="C://Users//artur//OneDrive//analytics//resources//")
        print(navs)
        options_risk_premia = options_risk_premia['1w'].rename('risk-premia')
        strategy_nav = navs['portfolio_nav'].rename('Short Straddle')

        fig1, fig2 = final_figures(options_risk_premia=options_risk_premia,
                                   strategy_nav=strategy_nav,
                                   rp_bound=5.0)
        qis.set_suptitle(fig=fig1, title='Risk-premia from weekly options', fontsize=16)
        qis.save_fig(fig=fig1, file_name='figure4_risk_premia_ts', local_path=local_path)

        qis.set_suptitle(fig=fig2, title='Performance of short straddle strategy', fontsize=16)
        qis.save_fig(fig=fig2, file_name='figure5_straddle', local_path=local_path)

    plt.show()


if __name__ == '__main__':

    unit_test = UnitTests.FINAL_FIGURES

    is_run_all_tests = False
    if is_run_all_tests:
        for unit_test in UnitTests:
            run_unit_test(unit_test=unit_test)
    else:
        run_unit_test(unit_test=unit_test)



"""
check mc pricer with kernel
"""
import numpy as np
import pandas as pd
import matplotlib.pyplot as plt
import qis as qis
from numba import njit
from scipy.special import betainc, gamma
from scipy.optimize import fsolve
from typing import Union
from enum import Enum

from stochvolmodels.pricers.analytic.tdist import compute_upsilon
from stochvolmodels import infer_bsm_ivols_from_slice_prices


# @njit
def generate_tvars_stock_path(nu: float = 4.5,
                              n_path: int = 10000,
                              ttm: float = 1.0 / 12.0,
                              vol: float = 0.2
                              ) -> np.ndarray:
    rv = np.random.standard_t(df=nu, size=n_path)
    upsilon = compute_upsilon(vol=vol, ttm=ttm, nu=nu)
    prices_t = 1.0 + upsilon*rv
    prices_t = prices_t + (1.0 - np.nanmean(prices_t))
    return prices_t


@njit
def compute_kernel(prices_t: np.ndarray, b: float,
                   nu: float = 4.5,
                   ttm: float = 1.0 / 12.0,
                   vol: float = 0.2
                   ) -> np.ndarray:
    """
    spot = 1 + x => x = spot - 1
    """
    x = prices_t - 1
    coeff = 3.0*ttm*vol*vol*(nu-2.0) / (nu-4.0)
    a = -b*coeff
    kernel = 1.0 + a*x+b*x*x*x
    kernel = np.where(kernel > 0.0, kernel, 0.0001)
    return kernel


def compute_implied_vols(b: float = -1,
                         nu: float = 4.5,
                         n_path: int = 100000,
                         ttm: float = 1.0 / 12.0,
                         vol: float = 0.2
                         ) -> pd.DataFrame:

    prices_t = generate_tvars_stock_path(nu=nu, n_path=n_path, ttm=ttm, vol=vol)
    kernel = compute_kernel(prices_t=prices_t, b=b, nu=nu, ttm=ttm, vol=vol)
    kernel = kernel / np.nanmean(kernel*prices_t)
    print(np.nanmean(kernel*prices_t))

    strikes = np.linspace(0.5, 1.5, 40)
    optiontypes = np.where(strikes < 1.0, 'P', 'C')

    model_prices = np.zeros_like(strikes)
    model_prices_kernel = np.zeros_like(strikes)
    for idx, (strike, type_) in enumerate(zip(strikes, optiontypes)):
        if type_ == 'C':
            payoff = np.where(np.greater(prices_t, strike), prices_t-strike, 0.0)
        else:
            payoff = np.where(np.less(prices_t, strike), strike - prices_t, 0.0)
        model_prices[idx] = np.nanmean(payoff)
        model_prices_kernel[idx] = np.nanmean(kernel*payoff)

    bsm_vols = infer_bsm_ivols_from_slice_prices(ttm=ttm, forward=1.0, strikes=strikes,
                                                 optiontypes=optiontypes,
                                                 model_prices=model_prices,
                                                 discfactor=1.0)
    print(model_prices_kernel)
    bsm_vols_kernel = infer_bsm_ivols_from_slice_prices(ttm=ttm, forward=1.0, strikes=strikes,
                                                        optiontypes=optiontypes,
                                                        model_prices=model_prices_kernel,
                                                        discfactor=1.0)

    bsm_vols = pd.Series(bsm_vols, index=strikes, name='T-vols')
    bsm_vols_kernel = pd.Series(bsm_vols_kernel, index=strikes, name='T-vols - kernel')
    vols = pd.concat([bsm_vols, bsm_vols_kernel], axis=1)

    return vols


bsm_vols = compute_implied_vols(b=-10.0,
                                nu=5.0,
                                ttm=1.0 / 12.0,
                                n_path=500000)

qis.plot_line(df=bsm_vols)

plt.show()



import numpy as np
import pandas as pd
import matplotlib.pyplot as plt
import seaborn as sns
from scipy import special as sps
from scipy.optimize import minimize
from scipy.stats import norm
from typing import Optional
from enum import Enum
import qis

from stochvolmodels.my_papers.volatility_models.load_data import fetch_ohlc_vol
from stochvolmodels import LogSvParams, HestonParams


def lognormal_sv_ss_pdf(sigma: np.ndarray,
                        params: LogSvParams
                        ) -> np.ndarray:
    nu = 2.0*(params.kappa2*params.theta-params.kappa1)/params.vartheta2 - 1.0
    q = 2.0*params.kappa1*params.theta/params.vartheta2
    b = 2.0*params.kappa2/params.vartheta2
    if params.kappa1 >= 1e-6:
        if params.kappa2 >= 1e-6:
            c = np.power(b/q, nu/2.0) / (2.0*sps.kv(nu, 2.0*np.sqrt(q*b)))
        else:
            c = np.power(q, -nu) / sps.gamma(-nu)
    else:
        raise NotImplementedError(f"kappa1 = 0 is not implemented")
    g = c*np.power(sigma, nu-1.0)*np.exp(-q*np.reciprocal(sigma)-b*sigma)
    return g


def lognormal_sv_ss_log_pdf(log_sigma: np.ndarray,
                            params: LogSvParams
                            ) -> np.ndarray:
    sigma = np.exp(log_sigma)
    return sigma*lognormal_sv_ss_pdf(sigma=sigma, params=params)


def heston_ss_pdf(var: np.ndarray,
                  params: HestonParams
                  ) -> np.ndarray:
    alpha = np.square(params.volvol) / (2.0*params.kappa)
    beta = params.theta / alpha
    c = np.power(alpha, beta) * sps.gamma(beta)
    g = np.power(var, beta-1.0)*np.exp(-var/alpha) / c
    return g


def heston_ss_log_vol_pdf(log_sigma: np.ndarray,
                          params: HestonParams
                          ) -> np.ndarray:
    var = np.exp(2.0*log_sigma)
    return 2.0*var*heston_ss_pdf(var=var, params=params)


def compute_vol_histogram(vol: pd.Series, bins: int = 100) -> pd.Series:
    # Get histogram of original data
    y, x = np.histogram(vol.to_numpy(), bins=bins, density=True)
    x = (x + np.roll(x, -1))[:-1] / 2.0
    hist = pd.Series(y / np.sum(y), index=x)
    return hist


def estimate_beta_sv():
    pass


def plot_estimated_beta_sv(vol: pd.Series,
                           bins: int = 50,
                           is_log: bool = True
                           ):
    hist = compute_vol_histogram(vol=vol, bins=bins)
    params = LogSvParams(theta=0.17, kappa1=2.0, kappa2=2.0, beta=-1.0, volvol=2.0)
    sigma = hist.index.to_numpy()
    dv = sigma[1] - sigma[0]
    analytic = lognormal_sv_ss_pdf(sigma=sigma, params=params)*dv
    with sns.axes_style("darkgrid"):
        fig, ax = plt.subplots(1, 1, figsize=(18, 10), tight_layout=True)

        if is_log:
            analytic = pd.Series(analytic, index=hist.index)
            qis.plot_line(df=analytic, ax=ax)
            qis.plot_line(hist, colors=['red'],
                          markers=["*"],
                          linewidth=0.0,
                          ax=ax)
            ax.set_yscale('log')
        else:
            hist.plot.bar(stacked=False, color='green', edgecolor='none', linewidth=0.0, width=1.0, ax=ax)
            current_ticks = ax.get_xticks()  # sigma' is mapped into ticks
            analytic = pd.Series(analytic, index=current_ticks)
            qis.plot_line(analytic, ax=ax)
            qis.reset_xticks(ax=ax, data=hist.index.to_numpy())


def plot_estimated_beta_log_sv(vol: pd.Series,
                               params: LogSvParams = LogSvParams(theta=0.14, kappa1=2.0, kappa2=2.0, beta=-1.0, volvol=1.0),
                               bins: int = 50,
                               is_log: bool = True
                               ):
    log_sigma = np.log(vol).replace([np.inf, -np.inf], np.nan).dropna()
    hist = compute_vol_histogram(vol=log_sigma, bins=bins)
    dv = hist.index[1] - hist.index[0]
    analytic = lognormal_sv_ss_log_pdf(log_sigma=hist.index.to_numpy(), params=params)*dv
    with sns.axes_style("darkgrid"):
        fig, ax = plt.subplots(1, 1, figsize=(18, 10), tight_layout=True)

        if is_log:
            analytic = pd.Series(analytic, index=hist.index)
            qis.plot_line(df=analytic, ax=ax)
            qis.plot_line(hist, colors=['red'],
                          markers=["*"],
                          linewidth=0.0,
                          ax=ax)
            ax.set_yscale('log')
        else:
            # hist.plot.bar(stacked=False, color='green', edgecolor='none', linewidth=0.0, width=1.0, ax=ax)
            analytic = pd.Series(analytic, index=hist.index, name='analytic')
            distr = pd.concat([hist, analytic], axis=1)
            qis.plot_line(distr, ax=ax)
            # qis.reset_xticks(ax=ax, data=hist.index.to_numpy())


def fit_distribution_log_sv(vol: pd.Series,
                            bins: int = 50
                            ) -> LogSvParams:
    log_sigma = np.log(vol).replace([np.inf, -np.inf], np.nan).dropna()
    hist = compute_vol_histogram(vol=log_sigma, bins=bins)
    dv = hist.index[1] - hist.index[0]

    def unpack_pars(pars: np.ndarray) -> LogSvParams:
        theta, kappa1, volvol = pars[0], pars[1], pars[2]
        params = LogSvParams(sigma0=theta, theta=theta, kappa1=kappa1, kappa2=kappa1, beta=0.0, volvol=volvol)
        return params

    def objective(pars: np.ndarray, args: np.ndarray) -> float:
        params = unpack_pars(pars=pars)
        analytic = lognormal_sv_ss_log_pdf(log_sigma=hist.index.to_numpy(), params=params) * dv
        sse = np.nansum(np.square(hist.to_numpy() - analytic))
        return sse

    options = {'disp': True, 'ftol': 1e-8}
    p0 = np.abs([0.15, 4.0, 1.0])
    bounds = ((0.05, 1.0), (0.5, 10), (0.1, 5.0))
    res = minimize(objective, p0, args=None, method='SLSQP', bounds=bounds, options=options)

    fit_params = unpack_pars(pars=res.x)
    return fit_params


def fit_distribution_log_sv_fixed_kappa(vol: pd.Series,
                                        kappa1: float,
                                        kappa2: float,
                                        bins: int = 50
                                        ) -> LogSvParams:
    """
    given kappa_1 and kappa_2 fit theta and volvol
    """
    log_sigma = np.log(vol).replace([np.inf, -np.inf], np.nan).dropna()
    hist = compute_vol_histogram(vol=log_sigma, bins=bins)
    dv = hist.index[1] - hist.index[0]

    def unpack_pars(pars: np.ndarray) -> LogSvParams:
        theta, volvol = pars[0], pars[1]
        params = LogSvParams(sigma0=theta, theta=theta, kappa1=kappa1, kappa2=kappa2, beta=0.0, volvol=volvol)
        return params

    def objective(pars: np.ndarray, args: np.ndarray) -> float:
        params = unpack_pars(pars=pars)
        analytic = lognormal_sv_ss_log_pdf(log_sigma=hist.index.to_numpy(), params=params) * dv
        sse = np.nansum(np.square(hist.to_numpy() - analytic))
        return sse

    options = {'disp': False, 'ftol': 1e-8}
    p0 = np.array([0.3, 1.0])
    bounds = ((0.05, 1.0), (0.25, 5.0))
    res = minimize(objective, p0, args=None, method='SLSQP', bounds=bounds, options=options)

    fit_params = unpack_pars(pars=res.x)
    return fit_params


def fit_distribution_heston(vol: pd.Series,
                            bins: int = 50
                            ) -> HestonParams:
    log_sigma = np.log(vol).replace([np.inf, -np.inf], np.nan).dropna()
    hist = compute_vol_histogram(vol=log_sigma, bins=bins)
    dv = hist.index[1] - hist.index[0]

    def unpack_pars(pars: np.ndarray) -> HestonParams:
        theta, kappa, volvol = pars[0], pars[1], pars[2]
        params = HestonParams(v0=theta, theta=theta, kappa=kappa, rho=0.0, volvol=volvol)
        return params

    def objective(pars: np.ndarray, args: np.ndarray) -> float:
        params = unpack_pars(pars=pars)
        analytic = heston_ss_log_vol_pdf(log_sigma=hist.index.to_numpy(), params=params) * dv
        sse = np.nansum(np.square(hist.to_numpy() - analytic))
        return sse

    options = {'disp': True, 'ftol': 1e-8}
    p0 = np.abs([0.04, 4.0, 1.0])
    bounds = ((0.001, 0.5), (0.5, 10), (0.1, 5.0))
    res = minimize(objective, p0, args=None, method='SLSQP', bounds=bounds, options=options)

    fit_params = unpack_pars(pars=res.x)
    return fit_params


def plot_estimated_svs(vol: pd.Series,
                       logsv_params: LogSvParams = LogSvParams(),
                       heston_params: Optional[HestonParams] = HestonParams(),
                       bins: int = 100,
                       ax: plt.Subplot = None,
                       **kwargs
                       ) -> None:
    log_sigma = np.log(vol).replace([np.inf, -np.inf], np.nan).dropna()
    hist = compute_vol_histogram(vol=log_sigma, bins=bins)
    hist_range = hist.index.to_numpy()

    dv = hist_range[1] - hist_range[0]
    analytic1 = pd.Series(lognormal_sv_ss_log_pdf(log_sigma=hist_range, params=logsv_params)*dv, index=hist_range, name='LogNormal')
    norm_mean, norm_std = np.mean(log_sigma), np.std(log_sigma)
    analytic3 = pd.Series(norm.pdf(x=hist_range, loc=norm_mean, scale=norm_std)*dv, index=hist_range, name='Normal PDF')

    if heston_params is not None:
        analytic2 = pd.Series(heston_ss_log_vol_pdf(log_sigma=hist_range, params=heston_params)*dv, index=hist_range, name='Heston')
        df = pd.concat([analytic1, analytic2, analytic3], axis=1)
    else:
        df = pd.concat([analytic1, analytic3], axis=1)

    if ax is None:
        fig, ax = plt.subplots(1, 1, figsize=(18, 10), tight_layout=True)

    qis.plot_line(hist, colors=['red'],
                  markers=["*"],
                  linewidth=0.0,
                  ax=ax,
                  **kwargs)

    colors = qis.get_n_colors(n=len(df.columns))
    qis.plot_line(df=df,
                  colors=colors,
                  y_limits=(1e-10, None),
                  xlabel='Log volatility',
                  xvar_format='{:.1f}',
                  ax=ax,
                  **kwargs)

    qis.set_legend(ax=ax,
                   markers=["*"]+["_"]*len(df.columns),
                   labels=['Empirical']+list(df.columns),
                   handlelength=0,
                   colors=['red']+colors,
                   legend_loc='lower center',
                   framealpha=0.90,
                   **kwargs)

    ax.set_yscale('log')


def produce_joint_report(vol: pd.Series, bins: int = 100):
    logsv_params = fit_distribution_log_sv(vol=vol, bins=bins)
    print(logsv_params)
    heston_params = fit_distribution_heston(vol=vol, bins=bins)
    print(heston_params)
    plot_estimated_svs(vol=vol, logsv_params=logsv_params, heston_params=heston_params, bins=bins)


class UnitTests(Enum):
    FETCH_VOL = 1
    PLOT_ESTIMATED_VOL = 2
    PLOT_ESTIMATED_LOG_VOL = 3
    FIT_LOGSV = 4
    FIT_HESTON = 5
    JOINT = 6


def run_unit_test(unit_test: UnitTests):

    if unit_test == UnitTests.FETCH_VOL:
        vol, returns = fetch_ohlc_vol(ticker='SPY')
        with sns.axes_style("darkgrid"):
            fig, ax = plt.subplots(1, 1, figsize=(18, 10), tight_layout=True)
            qis.plot_time_series(df=vol, ax=ax)
        hist = compute_vol_histogram(vol=vol)
        qis.plot_bars(hist)

    elif unit_test == UnitTests.PLOT_ESTIMATED_VOL:
        vol, returns = fetch_ohlc_vol(ticker='SPY')
        plot_estimated_beta_sv(vol=vol)

    elif unit_test == UnitTests.PLOT_ESTIMATED_LOG_VOL:
        vol, returns = fetch_ohlc_vol(ticker='SPY')
        plot_estimated_beta_log_sv(vol=vol)

    elif unit_test == UnitTests.FIT_LOGSV:
        vol, returns = fetch_ohlc_vol(ticker='SPY')
        fit_params = fit_distribution_log_sv(vol=vol)
        print(fit_params)
        plot_estimated_beta_log_sv(vol=vol, params=fit_params, is_log=True)
        plot_estimated_beta_log_sv(vol=vol, params=fit_params, is_log=False)

    elif unit_test == UnitTests.FIT_HESTON:
        vol, returns = fetch_ohlc_vol(ticker='SPY')
        fit_params = fit_distribution_heston(vol=vol)
        print(fit_params)

    elif unit_test == UnitTests.JOINT:
        vol, returns = fetch_ohlc_vol(ticker='MOVE')
        produce_joint_report(vol=vol)

    plt.show()


if __name__ == '__main__':

    unit_test = UnitTests.JOINT

    is_run_all_tests = False
    if is_run_all_tests:
        for unit_test in UnitTests:
            run_unit_test(unit_test=unit_test)
    else:
        run_unit_test(unit_test=unit_test)




"""
computation of il payoff under log sv
"""
import numpy as np
import matplotlib.pyplot as plt
from numba import njit
from enum import Enum

# stochvolmodels pricers
from stochvolmodels import (get_transform_var_grid,
                            vanilla_slice_pricer_with_mgf_grid,
                            digital_slice_pricer_with_mgf_grid,
                            compute_integration_weights,
                            ExpansionOrder,
                            get_expansion_n,
                            compute_logsv_a_mgf_grid,
                            LogSvParams)


def logsv_il_pricer(params: LogSvParams,
                    ttm: float,
                    p1: float,  # can price on range of forwards
                    p0: float,
                    pa: float,
                    pb: float,
                    is_stiff_solver: bool = False,
                    expansion_order: ExpansionOrder = ExpansionOrder.SECOND,
                    vol_scaler: float = None,
                    notional: float = 1000000
                    ) -> float:
    """
    price il using bsm mgf
    """
    # starting values
    if vol_scaler is None:  # for calibrations we fix one vol_scaler so the grid is not affected by v0
        vol_scaler = params.sigma0 * np.sqrt(np.minimum(np.min(ttm), 0.5 / 12.0))

    # for vanilla call and put
    phi_grid, psi_grid, theta_grid = get_transform_var_grid(vol_scaler=vol_scaler,
                                                                 real_phi=-0.4)
    a_t0 = np.zeros((phi_grid.shape[0], get_expansion_n(expansion_order)), dtype=np.complex128)

    a_t0, log_mgf_grid = compute_logsv_a_mgf_grid(ttm=ttm,
                                                  phi_grid=phi_grid,
                                                  psi_grid=psi_grid,
                                                  theta_grid=theta_grid,
                                                  a_t0=a_t0,
                                                  expansion_order=expansion_order,
                                                  is_stiff_solver=is_stiff_solver,
                                                  **params.to_dict())

    vanilla_option_prices = vanilla_slice_pricer_with_mgf_grid(log_mgf_grid=log_mgf_grid,
                                                               phi_grid=phi_grid,
                                                               forward=p1,
                                                               strikes=np.array([pa, pb]),
                                                               optiontypes=np.array(['P', 'C']),
                                                               discfactor=1.0)
    bsm_put, bsm_call = vanilla_option_prices[0], vanilla_option_prices[1]

    digital_options = digital_slice_pricer_with_mgf_grid(log_mgf_grid=log_mgf_grid,
                                                         phi_grid=phi_grid,
                                                         forward=p1,
                                                         strikes=np.array([pa, pb]),
                                                         optiontypes=np.array(['P', 'C']),
                                                         discfactor=1.0)
    digital_put, digital_call = digital_options[0], digital_options[1]

    square_root = square_root_payoff_pricer_with_mgf_grid(log_mgf_grid=log_mgf_grid,
                                                          phi_grid=phi_grid,
                                                          forward=p1,
                                                          pa=pa,
                                                          pb=pb,
                                                          discfactor=1.0)

    sp0 = np.sqrt(p0)
    spa = np.sqrt(pa)
    spb = np.sqrt(pb)

    linear = sp0*(p1/p0+1.0)

    payoff = -2.0*square_root + linear + \
             (1.0 / spa) * bsm_put - (1.0 / spb) * bsm_call \
             -2.0 * spa * digital_put -2.0 * spb * digital_call

    notional0 = 1.0 / (2.0*sp0-p0/spb-spa)
    payoff = - (notional0*notional)*payoff
    return payoff


logsv_il_pricer_vector = np.vectorize(logsv_il_pricer, doc='Vectorized `logsv_il_pricer`')


@njit(cache=False, fastmath=True)
def square_root_payoff_pricer_with_mgf_grid(log_mgf_grid: np.ndarray,
                                            phi_grid: np.ndarray,
                                            forward: float,
                                            pa: float,
                                            pb: float,
                                            discfactor: float = 1.0,
                                            is_simpson: bool = True
                                            ) -> np.ndarray:
    """
    generic function for pricing digital options on the spot given the mgf grid
    mgf in x is function defined on log-price transform phi grids
    transform variable is phi_grid = real_phi + i*p
    grid can be non-uniform
    we can use either positive or negative phi_real but not
    """
    dp = compute_integration_weights(var_grid=phi_grid, is_simpson=is_simpson)

    x = np.log(forward)
    xa = np.log(pa)
    xb = np.log(pb)
    p_payoff = (np.exp( (phi_grid+0.5)*xb - phi_grid*x) - np.exp((phi_grid+0.5)*xa - phi_grid*x))
    p_payoff = (dp / np.pi) * p_payoff / (phi_grid+0.5)
    option_price = discfactor * np.nansum(np.real(p_payoff*np.exp(log_mgf_grid)))
    return option_price


class UnitTests(Enum):
    COMPUTE_MODEL_PRICES = 1


def run_unit_test(unit_test: UnitTests):

    # define model params
    params = LogSvParams(sigma0=0.4861785891939535, theta=0.6176006871606874, kappa1=1.955809653686808, kappa2=1.978367101612294, beta=-0.26916969112829325, volvol=3.265815229306317)

    if unit_test == UnitTests.COMPUTE_MODEL_PRICES:
        payoff = logsv_il_pricer(params=params,
                                 ttm=10.0/365.0,
                                 p1=2200.0,
                                 p0=2200.0,
                                 pa=2000.0,
                                 pb=2400.0)
        print(payoff)

    plt.show()


if __name__ == '__main__':

    unit_test = UnitTests.COMPUTE_MODEL_PRICES

    is_run_all_tests = False
    if is_run_all_tests:
        for unit_test in UnitTests:
            run_unit_test(unit_test=unit_test)
    else:
        run_unit_test(unit_test=unit_test)



import string
import numpy as np
import pandas as pd
import matplotlib.pyplot as plt
import qis
import seaborn as sns
from typing import Dict, List
from enum import Enum

# package
from stochvolmodels.pricers.logsv_pricer import LogSvParams
from stochvolmodels.utils.funcs import set_seed

# project
from stochvolmodels.my_papers.volatility_models.load_data import fetch_ohlc_vol
import stochvolmodels.my_papers.volatility_models.ss_distribution_fit as ssd
from stochvolmodels.my_papers.volatility_models.vol_beta import estimate_vol_beta
from stochvolmodels.my_papers.volatility_models.autocorr_fit import autocorr_fit_report_logsv

KWARGS = dict(fontsize=14)
FIGSIZE = (18, 8)


def plot_vols(tickers: List[str]) -> plt.Figure:

    vols = {}
    for ticker in tickers:
        vol, returns = fetch_ohlc_vol(ticker=ticker)
        vols[ticker] = vol
    vols = pd.DataFrame.from_dict(vols, orient='columns')

    with sns.axes_style("darkgrid"):
        fig, axs = plt.subplots(1, 2, figsize=FIGSIZE, tight_layout=True)

        qis.plot_time_series(df=vols,
                             x_date_freq='A',
                             title="(A) Time Series",
                             var_format='{:,.0%}',
                             legend_loc='upper center',
                             legend_stats=qis.LegendStats.AVG_NONNAN_LAST,
                             y_limits=(0.0, None),
                             ax=axs[0],
                             **KWARGS)

        qis.plot_histogram(vols,
                           xlabel='Vol',
                           title="(B) Empirical PDF",
                           legend_loc='upper center',
                           xvar_format='{:,.0%}',
                           desc_table_type=qis.DescTableType.NONE,
                           x_limits=(0.0, 2.5),
                           ax=axs[1],
                           **KWARGS)

    return fig


def plot_autocorrs(model_params: Dict[str, LogSvParams],
                   nb_path: int = 5000,
                   num_lags: int = 120,
                   ttm: float = 10.0
                   ) -> plt.Figure:

    with sns.axes_style("darkgrid"):
        fig, axs = plt.subplots(1, len(model_params.keys()), figsize=FIGSIZE, tight_layout=True)

        for idx, (ticker, logsv_params) in enumerate(model_params.items()):
            vol, returns = fetch_ohlc_vol(ticker=ticker)
            autocorr_fit_report_logsv(vol=vol, params=logsv_params, nb_path=nb_path, num_lags=num_lags, ttm=ttm,
                                      ax=axs[idx],
                                      **KWARGS)
            qis.set_title(ax=axs[idx], title=f"{string.ascii_uppercase[idx]}) {ticker}", **KWARGS)
        qis.align_y_limits_axs(axs)

    return fig


def plot_ss_distributions(model_params: Dict[str, LogSvParams],
                          bins: int = 50
                          ) -> plt.Figure:

    with sns.axes_style("darkgrid"):
        if len(model_params.keys()) == 4:
            fig, axs = plt.subplots(2, 2, figsize=(18, 14), tight_layout=True)
            axs = qis.to_flat_list(axs)
        else:
            fig, axs = plt.subplots(1, len(model_params.keys()), figsize=FIGSIZE, tight_layout=True)

        for idx, (ticker, logsv_params) in enumerate(model_params.items()):
            vol, returns = fetch_ohlc_vol(ticker=ticker)
            heston_params = ssd.fit_distribution_heston(vol=vol, bins=bins)
            ssd.plot_estimated_svs(vol=vol, logsv_params=logsv_params, heston_params=heston_params, bins=bins,
                                   ax=axs[idx],
                                   **KWARGS)
            qis.set_title(ax=axs[idx], title=f"{string.ascii_uppercase[idx]}) {ticker}", **KWARGS)
        qis.align_y_limits_axs(axs)

    return fig


def vol_beta_plots(tickers: List[str], span: int = 65) -> plt.Figure:

    vol_betas = {}
    for idx, ticker in enumerate(tickers):
        vol, returns = fetch_ohlc_vol(ticker=ticker)
        vol_betas[ticker] = estimate_vol_beta(vol=vol, returns=returns, span=span)
    vol_betas = pd.DataFrame.from_dict(vol_betas, orient='columns')

    with sns.axes_style("darkgrid"):
        fig, axs = plt.subplots(1, 2, figsize=FIGSIZE, tight_layout=True)
        qis.plot_time_series(vol_betas,
                             x_date_freq='A',
                             legend_stats=qis.LegendStats.AVG_NONNAN_LAST,
                             trend_line=qis.TrendLine.ABOVE_ZERO_SHADOWS,
                             title="(A) Time series",
                             date_format='%d-%b-%y',
                             legend_loc='upper left',
                             framealpha=0.75,
                             ax=axs[0],
                             **KWARGS)

        # fig, ax = plt.subplots(1, 1, figsize=FIGSIZE, tight_layout=True)
        qis.plot_histogram(vol_betas,
                           xlabel='Vol beta',
                           title="(B) Empirical PDF",
                           legend_loc='upper center',
                           desc_table_type=qis.DescTableType.NONE,
                           ax=axs[1],
                           **KWARGS)

    return fig


class UnitTests(Enum):
    PLOT_VOLS = 1
    AUTOCORRELATION_PLOTS = 2
    SS_DENSITY_PLOTS = 3
    VOL_BETA_PLOTS = 4
    MODEL_PARAMS_TABLE = 5


def run_unit_test(unit_test: UnitTests):

    set_seed(3)
    np.random.seed(3)

    vix_log_params = LogSvParams(sigma0=0.19928505844247962, theta=0.19928505844247962, kappa1=1.2878835150774184,
                                 kappa2=1.9267876555824357, beta=0.0, volvol=0.7210463316739526)

    move_log_params = LogSvParams(sigma0=0.9109917133860931, theta=0.9109917133860931, kappa1=0.1,
                                  kappa2=0.41131244621275886, beta=0.0, volvol=0.3564212939473691)

    ovx_log_params = LogSvParams(sigma0=0.3852514800317871, theta=0.3852514800317871, kappa1=2.7774564907918027,
                                 kappa2=2.2351296851221107, beta=0.0, volvol=0.8344408577025486)

    btc_log_params = LogSvParams(sigma0=0.7118361434192538, theta=0.7118361434192538,
                                 kappa1=2.214702576955766, kappa2=2.18028273418397, beta=0.0, volvol=0.921487415907961)

    eth_log_params = LogSvParams(sigma0=0.8657438901704476, theta=0.8657438901704476, kappa1=1.955809653686808,
                                 kappa2=1.978367101612294, beta=0.0, volvol=0.8484117641903834)

    model_params = {'VIX': vix_log_params,
                    'MOVE': move_log_params,
                    'OVX': ovx_log_params,
                    'BTC': btc_log_params}

    local_path = "C://Users//artur//OneDrive//My Papers//Working Papers//Review of Beta Lognormal SV Model. Zurich. Nov 2023//figures//"

    if unit_test == UnitTests.PLOT_VOLS:
        fig = plot_vols(tickers=list(model_params.keys()))
        is_save = True
        if is_save:
            qis.save_fig(fig, file_name='vols_ts', local_path=local_path)

    elif unit_test == UnitTests.AUTOCORRELATION_PLOTS:
        fig = plot_autocorrs(model_params=model_params,
                             nb_path=10000, num_lags=120, ttm=10.0)
        is_save = True
        if is_save:
            qis.save_fig(fig, file_name='autocorr_fit', local_path=local_path)

    elif unit_test == UnitTests.SS_DENSITY_PLOTS:
        fig = plot_ss_distributions(model_params=model_params)
        is_save = True
        if is_save:
            qis.save_fig(fig, file_name='ss_distribution', local_path=local_path)

    elif unit_test == UnitTests.VOL_BETA_PLOTS:
        tickers = ['VIX', 'MOVE', 'OVX', 'BTC']
        fig = vol_beta_plots(tickers=tickers)
        is_save = True
        if is_save:
            qis.save_fig(fig, file_name='vol_beta', local_path=local_path)

    elif unit_test == UnitTests.MODEL_PARAMS_TABLE:
        data = {ticker: model_param.to_dict() for ticker, model_param in model_params.items() }
        df = pd.DataFrame.from_dict(data, orient='columns')
        df = df.drop(['sigma0', 'beta'], axis=0)
        dfs = qis.df_to_str(df=df)
        print(dfs)
        text = dfs.to_latex()
        print(text)

    plt.show()


if __name__ == '__main__':

    unit_test = UnitTests.MODEL_PARAMS_TABLE

    is_run_all_tests = False
    if is_run_all_tests:
        for unit_test in UnitTests:
            run_unit_test(unit_test=unit_test)
    else:
        run_unit_test(unit_test=unit_test)
